
/**
 * Scrowser - a server-side scraping browser
 */

var http = require("http");
var url = require("url");
var sys = require("sys");
var events = require("events");
var jsdom = require('jsdom');  // Get the DOM, server-side
var fs = require('fs');
var rest = require('restler');  // Used for uploading files

// This is a hack to get cookies to work with http library headers. this will likely break any other libraries using node.js/http
// As far as i can see, node.js is broken how it handles headers and this change should be implemented in node.js core
http.IncomingMessage.prototype._addHeaderLine = function (field, value) {
	if(field == "set-cookie") {
		if (this.headers.hasOwnProperty(field)) {
			this.headers[field].push(value);
		} else {
			this.headers[field] = [value];
		}
	}
	else {
		if (this.headers.hasOwnProperty(field)) {
			this.headers[field] += ", " + value;
		} else {
			this.headers[field] = value;
		}
	}
};

try {
	var compress = require("./compress");
}
catch(err) {
	if( err.message.indexOf("Cannot find module") >= 0 ) {
		sys.puts("no compress");
		var compress = null;
	}
	else {
		throw err;
	}
}

function scrowser() {
	var cookies = [];
        var me = this;
	
	if( compress !== null ) {
		var gunzip = new compress.Gunzip;
	}
	
	var clients = {
	};

	this.clients = clients;
	
	this.perform = function(rurl, method, cb, data, exheaders, tlscb) {
                method = method.toString().toUpperCase();
		this.clientheaders = exheaders;
		var curl = url.parse(rurl);
		var key = curl.protocol + "//" + curl.hostname;
		
		if(!clients[key]) {
			var client = null;
			switch(curl.protocol) {
				case "https:":
					client = http.createClient(curl.port||443, curl.hostname);
					client.setSecure();
//"x509_PEM");
				break;
				default:
					client = http.createClient(curl.port||80, curl.hostname);
					break;
			}
			clients[key] = {
				"http": client,
				"headers": {}
			};
		}
		
		clients[key].headers = {
			"User-Agent": this.getUserAgent(),
			"Accept" : "*/*",
			"Connection" : "close",
			"Host" : curl.hostname
		};
		if(method == "POST") {
			clients[key].headers["Content-Length"] = data.length;
			clients[key].headers["Content-Type"] = "application/x-www-form-urlencoded";
		}
		for (attr in exheaders) {clients[key].headers[attr] = exheaders[attr];}		
		if(!compress && clients[key].headers["Accept-Encoding"]) {
			clients[key].headers["Accept-Encoding"] = "none";
		}
		var mycookies = this.filterCookies(rurl);

		if( mycookies.length > 0 ) {
			clients[key].headers["Cookie"] = mycookies.join(";");
		}
		
		var target = "";
		if(curl.pathname) target += curl.pathname;
		if(curl.search) target += curl.search;
		if(curl.hash) target += curl.hash;
		if(target=="") target = "/";

		var req = clients[key].http.request(method, target, clients[key].headers);
		req.addListener("response", function(res) {
			var mybody = [];
			if(tlscb) {
				if(!tlscb({
						"status" : res.connection.verifyPeer(), 
						"certificate" : res.connection.getPeerCertificate("DNstring")
					}
				)) {
					cb(-2, null, null);		
					return;
				}
			}
			res.setEncoding("utf8");
			if(res.headers["content-encoding"] == "gzip") {
				res.setEncoding("binary");
			}
			res.addListener("data", function(chunk) {
				mybody.push(chunk);
			});
			res.addListener("end", function() {
				var body = mybody.join("");
				if( compress !== null && res.headers["content-encoding"] == "gzip") {
					gunzip.init();
					body = gunzip.inflate(body, "binary");
					gunzip.end();
				}
				var resp = {
					"response": {
						"status" : res.statusCode,
						"headers" : res.headers,
						"body-length" : body.length,
						"body" : body
					},
					"request": {
						"url" : rurl,
						"headers" : clients[key].headers
					}
				}
				cb(resp);	
			});
			res.addListener("error", function() {
				cb(-1, res.headers, mybody.join(""));		
			});
			if(res.headers["set-cookie"]) {
				res.headers["set-cookie"].forEach( function( cookie ) {
                                    me.setCookie(me.parseCookieString(cookie));
				});
			}
		});
		if(method == "POST") {
			req.write(data);
		}
		req.end();
	}

	this.getCookie = function(name, domain) {
                if (!domain) domain = this.getHost();
		var mycookies = cookies.filter(function(value, index, arr) {
			return(domain == value.domain || domain == 'www' + value.domain || domain == 'www.' + value.domain);
		});
		for( var i=0; i < mycookies.length; i++ ) {
			parts = mycookies[i].value.split("=");
			if( parts[0] == name ) {
				return parts[1];
			}
		}
		return null;
	}

        this.filterCookies = function(url_arg) {
            var curl = url.parse(url_arg),
                mycookies = [];
            cookies.filter(function(value, index, arr) {
                return(curl.hostname.substring(curl.hostname.length - value.domain.length) == value.domain);
            }).forEach( function(cookie) {
                    mycookies.push(cookie.value);
            });
            return mycookies;
        }
        
        this.parseCookieString = function(cookie) {
            var props = cookie.split(";");
            var newcookie = {
                    "value": "",
                    "domain": "",
                    "path": "/",
                    "expires": ""
            };
            newcookie.value = props.shift();
            props.forEach( function( prop ) {
                    var parts = prop.split("="),
                    name = parts[0].trim();
                    switch(name.toLowerCase()) {
                            case "domain":
                                    newcookie.domain = parts[1].trim();
                                    break;
                            case "path":
                                    newcookie.path = parts[1].trim();
                                    break;
                            case "expires":
                                    newcookie.expires = parts[1].trim();
                                    break;
                    }
            });
            return newcookie;
        }

        /**
            var newcookie = {
                "value": "name=value",
                "domain": "www.somedomain.com",
                "path": "/",
                "expires": ""
            };
         */
        this.setCookie = function(newcookie) {
            if(newcookie.domain == "") newcookie.domain = this.getHost();
            var match = cookies.filter(function(value, index, arr) {
                    if(value.domain == newcookie.domain && value.path == newcookie.path && value.value.split("=")[0] == newcookie.value.split("=")[0]) {
                            arr[index] = newcookie;  // update cookie
                            return true;
                    }
                    else {
                            return false;
                    }
            });
            if(match.length == 0) {
                cookies.push(newcookie);  // add new cookie
                me.storeCookies();  // persist cookies
            }
        }
	
	this.end = function() {
		clients.forEach(function(client) {
			client.http.end();		
		});
	}

        /**
         * MORE
         */
    // Manage the browser id. Used to identify the browser between node.js requests.
    this.browserId = 'default';
    this.setBrowserId = function(browserId) {
        this.browserId = browserId;
    }
    this.getBrowserId = function() {
        return this.browserId;
    }
    // Current host
    this.host = '';
    this.getHost = function() {
        return this.host;
    }
    this.setHost = function(host) {
        this.host = host;
    }

    /**
     * Event called when an HTML document has been retrieved.
     * Set to false to disable.
     * Stops program execution when errors are encountered
     * @param {jQuery} $ A jQuery object containing a parsed HTML document. This will be FALSE when no DOM is available
     * @param {object} o An object containing the complete HTTP request and response
     * @param {function} cb Callback
     */
    /*
    this.onGetHtml = function($, o, cb) {
        cb.call(this, $, o);
    }
    */
    this.onGetHtml = false;

    /**
     * Handle errors
     * @param {scrowser_error} errors A collection of error(s) and related information.
     */
    this.onGetHtmlError = function(errors) {
        console.log(errors);
        throw new Error("onGetHtml() - errors encountered")
    }

    /**
     * Get a response that contains an HTML document, loads the DOM and initalizes jquery for traversal.
     * Persists cookies and maintains referer information, so subsequent calls to getHeaders() will pass updated header info
     * @param {string} rurl The URL to load relative to the current host. When a protocol is used, ignores the current host. When empty goes to the index document of the current host
     * @param {string} method POST or GET
     * @param {function} callback Called on a succesful HTTP request/response cycle. See onGetHtml() for a description of all arguments passed to the callback. errors are not passed to the callback - instead, errors stop script execution before the callback can be triggered
     * @param {data} string A string of URL-encoded data to send with a POST. Ignored on GET.
     */
    this.getHtml = function(rurl, method, callback, data) {
         var rurl2 = this.normalizeUrl(rurl);
         this.perform(rurl2, method, function(o, error_headers, error_body) {
            if (o == -1) {
                me.onGetHtmlError(new scrowser_error(
                    'An error occured while accessing "' + rurl2 + '"',
                    scrowser_error.HTTP,
                    error_headers,
                    error_body));
            } else {
                jsdom.env(me.removeEmptyScriptTags(o.response.body), [ 'http://code.jquery.com/jquery-1.5.min.js' ], function(errors, window) {
                    if (errors) {
                        console.log('Unable to parse document at "'+rurl2+'"');
                        me.onGetHtmlError(new scrowser_error(
                            errors,
                            scrowser_error.DOM,
                            o.response.headers,
                            o.response.body));
                    } else {
                        if (me.onGetHtml) {
                            me.onGetHtml(window.$, o, callback);
                        } else {
                            callback.call(me, window.$, o);
                        }
                    }
                });
            }
         }, data);
    }

    // In some cirumstances empty <SCRIPT> tags combined with jQuery and jsdom cause jsdom.env() to quit without warning or error.
    this.removeEmptyScriptTags = function(body) {
        return body.replace(/<script[^>]*><\/script>/gim, '');
    }

    this.normalizeUrl = function(rurl) {
         // Normalize the url
         var rurl2;
         if (rurl) {
            var curl = url.parse(rurl);
            if (curl.protocol) {
                rurl2 = rurl;  // When a protocol is defined, used the full original url
            } else {
                rurl2 = 'http://' + this.getHost() + '/' + (rurl.indexOf('/') === 0 ? rurl.substr(1) : rurl);
            }
         } else {
             rurl2 = 'http://' + this.getHost();
         }
         return rurl2;
    }

    // Directory where cookies should be stored
    this.cookiesDir = '/tmp/scrowser_cookies';

    // Create the  the cookie directory. This is syncronous/blocking!
    this.initCookiesDir = function() {
        var dir;
        try {
            dir = fs.lstatSync(this.cookiesDir);
        } catch (e) {
            //console.log('Unable to stat cookies directory: ', e);
            dir = false;
        }
        if (dir && !dir.isDirectory()) {
            throw new Error(this.cookiesDir + ' is not a directory!')
        }
        if (dir) {
            return true;
        } else {
            fs.mkdirSync(this.cookiesDir, 0777);
        }
        return false;
    }

    // Get a path to a cookie storage file based on the current browser_id
    this.getCookiePath = function() {
        return this.cookiesDir + '/' + this.getBrowserId();
    }

    /**
     * Persist cookies to disk for the current browser_id
     * This is syncronous/blocking
     * Saves the cookies as a JSON-encoded string.
     */
    this.storeCookies = function() {
        this.initCookiesDir();
        var path = this.getCookiePath();
        var cs = JSON.stringify(cookies);
        fs.writeFileSync(path, cs);
    }

    /**
     * Load cookies from storage for the current browser_id
     * This is syncronous/blocking
     */
    this.loadCookies = function() {
        this.initCookiesDir();
        var path = this.getCookiePath();
        try {
            var cs = fs.readFileSync(path);
            cookies = JSON.parse(cs);
        } catch (e) {
            cookies = [];
            console.log('Unable to read cookie file at "'+path+'"');
        }
    }

    /**
     * Clear all locally stored cookies
     */
    this.clearCookies = function() {
        cookies = [];
        this.storeCookies();
    }

    /**
     * Populate a jQuery FORM node.
     * @param {jQuery} $
     * @param {jQuery node} form
     * @param {object} map Properties are the names of the form fields, property values are the values to to set matching form fields to
     */
    this.populateForm = function($, form, map) {
        var name;
        // Populate text, password, hidden, select and textarea fields
        $('INPUT[type=text], INPUT[type=hidden], INPUT[type=password], SELECT, TEXTAREA', form).each(function() {
            name = $(this).attr('name');
            if (name && map[name]) {
                $(this).val(map[name]);
            }
        });
        // Select checkbox and radios
        $('INPUT[type=checkbox], INPUT[type=radio]', form).each(function() {
            name = $(this).attr('name');
            if (name && map[name] && $(this).val() == map[name]) {
                $(this).attr('checked', true);
            }
        });
    }

    /**
     * Return a form as an hash whose keys are form element names and values are those element's values.
     * @param {jQuery} form
     */
    this.getFormHash = function(form) {
        var a = form.serializeArray();
        var o = {};
        for (var i in a) {
            o[a[i].name] = a[i].value;
        }
        return o;
    }
    
    /**
     * Upload a file
     * @param {string} file The full path to the file to upload
     * @param {string} file_post_name The name of the POST variable to bind the file to.
     * @param {string} furl The URL to POST the file to
     * @param {object} data Additional POST data.
     * @param {function} cb Callback is passed two args: a jQuery object of the DOM returned and the request/response headers - the second arg is always FALSE
     */
    this.uploadFile = function(file, file_post_name, furl, data, cb) {
        //var file = '/Users/lee/Downloads/ParallelsDesktop-parallels-en_US-6.0.11994.637263.dmg';
        var stat = require('fs').statSync(file);
        var size = stat['size'];
        if (!data) data = {};
        data[file_post_name] = rest.file(file, file, size);
        rest.post(this.normalizeUrl(furl), {
            multipart: true,
            data: data,
            headers: this.getHeaders(furl)
        }).on('complete', function(body) {
            jsdom.env(me.removeEmptyScriptTags(body), [ 'http://code.jquery.com/jquery-1.5.min.js' ], function(errors, window) {
                if (me.onGetHtml) {
                    me.onGetHtml(window.$, false, cb);
                } else {
                    cb.call(me, window.$, false);
                }
            });
        });
    }

    this.getHeaders = function(rurl) {
        var h = {},
            hurl = this.normalizeUrl(rurl);
        h["User-Agent"] = this.getUserAgent();
        var mycookies = this.filterCookies(hurl);
        if (mycookies.length > 0) {
            h["Cookie"] = mycookies.join(';');
        }
        return h;
    }

    this.getUserAgent = function() {
        return "Mozilla/5.0 (Macintosh; U; Intel Mac OS X 10.6; en-US; rv:1.9.2.16) Gecko/20110319 Firefox/3.6.16";
    }

}

/**
 * Mature error definition for scrowser
 * @param {string|array} errors Any error message(s) to report
 * @param {int} type One of the scrowser_error constants
 * @param {object} error_headers Headers returned by the related HTTP request
 * @param {string} error_body The body of the HTTP request.
 */
function scrowser_error(errors, type, error_headers, error_body) {
    this.type = type || scrowser_error.UNKNOWN;
    // errors must be an array
    this.errors = ((typeof errors != 'undefined' && typeof errors.prototype != 'undefined') ? (errors.prototype.toString.call(o) != '[object Array]' ? [ errors ] : errors) : false) || [];
    this.error_headers = error_headers;
    this.error_body = error_body;
}
scrowser_error.HTTP = 1; // A problem with an HTTP request/response cycle
scrowser_error.DOM = 2;  // A problem when parsing a document
scrowser_error.UNKNOWN = 3;  // An unknown error occured

exports.scrowser = scrowser;
exports.error = scrowser_error;
